# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Configuration file parser"

# pylint: disable=useless-object-inheritance

import logging
import os.path as p
import re
from glob import glob
from threading import RLock
from typing import Any, Dict, Iterable, List, Tuple, Union

from hdl_checker import exceptions
from hdl_checker.path import Path
from hdl_checker.types import BuildFlags, BuildFlagScope, FileType

# pylint: disable=invalid-name
_splitAtWhitespaces = re.compile(r"\s+").split
_replaceCfgComments = re.compile(r"(\s*#.*|\n)").sub
_configFileScan = re.compile(
    "|".join(
        [
            r"^\s*(?P<parameter>\w+)\s*(\[(?P<parm_lang>vhdl|verilog|systemverilog)\]|\s)*=\s*(?P<value>.+)\s*$",
            r"^\s*(?P<lang>(vhdl|verilog|systemverilog))\s+"
            r"(?P<library>\w+)\s+"
            r"(?P<path>[^\s]+)\s*(?P<flags>.*)\s*",
        ]
    ),
    flags=re.I,
).finditer
# pylint: enable=invalid-name


def _extractSet(entry):  # type: (str) -> BuildFlags
    """
    Extract a list by splitting a string at whitespaces, removing
    empty values caused by leading/trailing/multiple whitespaces
    """
    string = str(entry).strip()
    # Return an empty list if the string is empty
    if not string:
        return ()
    return tuple(_splitAtWhitespaces(string))


class ConfigParser(object):
    """
    Configuration info provider
    """

    _list_parms = {
        "single_build_flags": BuildFlagScope.single,
        "global_build_flags": BuildFlagScope.all,
        "dependencies_build_flags": BuildFlagScope.dependencies,
    }

    _single_value_parms = ("builder",)
    _deprecated_parameters = ("target_dir",)

    _logger = logging.getLogger(__name__ + ".ConfigParser")

    def __init__(self, filename):  # type: (Path) -> None
        self._logger.debug("Creating config parser for filename '%s'", filename)

        self._parms = {"builder": None}  # type: Dict[str, Union[str, None]]

        self._flags = {
            FileType.vhdl: {
                BuildFlagScope.single: (),
                BuildFlagScope.all: (),
                BuildFlagScope.dependencies: (),
            },
            FileType.verilog: {
                BuildFlagScope.single: (),
                BuildFlagScope.all: (),
                BuildFlagScope.dependencies: (),
            },
            FileType.systemverilog: {
                BuildFlagScope.single: (),
                BuildFlagScope.all: (),
                BuildFlagScope.dependencies: (),
            },
        }  # type: Dict[FileType, Dict[BuildFlagScope, BuildFlags] ]

        self._sources = []  # type: List[Tuple[str, str, BuildFlags]]

        self.filename = filename

        self._timestamp = 0.0
        self._parse_lock = RLock()

    def _shouldParse(self):  # type: () -> bool
        """
        Checks if we should parse the configuration file
        """
        return self.filename.mtime > self._timestamp

    def _updateTimestamp(self):
        # type: (...) -> Any
        """
        Updates our timestamp with the configuration file
        """
        self._timestamp = self.filename.mtime

    def isParsing(self):  # type: () -> bool
        "Checks if parsing is ongoing in another thread"
        locked = not self._parse_lock.acquire(False)
        if not locked:
            self._parse_lock.release()
        return locked

    def _parseIfNeeded(self):
        # type: () -> None
        """
        Locks accesses to parsed attributes and parses the configuration file
        """
        with self._parse_lock:
            if self._shouldParse():
                self._parse()

    def _parse(self):  # type: () -> None
        """
        Parse the configuration file without any previous checking
        """
        self._logger.info("Parsing '%s'", self.filename)
        self._updateTimestamp()
        self._sources = []
        for _line in open(self.filename.name, mode="rb").readlines():
            line = _replaceCfgComments("", _line.decode(errors="ignore"))
            self._parseLine(line)

    def _parseLine(self, line):  # type: (str) -> None
        """
        Parses a line a calls the appropriate extraction methods
        """
        for match in _configFileScan(line):
            groupdict = match.groupdict()
            self._logger.debug("match: '%s'", groupdict)
            if groupdict["parameter"] is not None:
                self._handleParsedParameter(
                    groupdict["parameter"], groupdict["parm_lang"], groupdict["value"]
                )
            else:
                for source_path in self._getSourcePaths(groupdict["path"]):
                    self._sources.append(
                        (
                            source_path,
                            {
                                "library": groupdict["library"],
                                "flags": _extractSet(groupdict["flags"]),
                            },
                        )
                    )

    def _handleParsedParameter(self, parameter, lang, value):
        # type: (str, str, str) -> None
        """
        Handles a parsed line that sets a parameter
        """
        self._logger.debug(
            "Found parameter '%s' for '%s' with value '%s'", parameter, lang, value
        )
        if parameter in self._deprecated_parameters:
            self._logger.debug("Ignoring deprecated parameter '%s'", parameter)
        elif parameter in self._single_value_parms:
            self._logger.debug("Handling '%s' as a single value", parameter)
            self._parms[parameter] = value
        elif parameter in self._list_parms:
            self._logger.debug("Handling '%s' as a list of values", parameter)
            self._flags[FileType(lang)][self._list_parms[parameter]] = _extractSet(
                value
            )
        else:
            raise exceptions.UnknownParameterError(parameter)

    def _getSourcePaths(self, path):  # type: (str) -> Iterable[str]
        """
        Normalizes and handles absolute/relative paths
        """
        source_path = p.normpath(p.expanduser(path))
        # If the path to the source file was not absolute, we assume
        # it was relative to the config file base path
        if not p.isabs(source_path):
            fname_base_dir = p.dirname(self.filename.abspath)
            source_path = p.join(fname_base_dir, source_path)

        return glob(source_path) or [source_path]

    def parse(self):
        # type: (...) -> Dict[Any, Any]
        """
        Parses the file if it hasn't been parsed before or if the config file
        has been changed
        """
        self._parseIfNeeded()
        data = {"sources": self._sources}  # type: Dict[Any, Any]

        builder_name = self._parms.get("builder", None)
        if builder_name is not None:
            data["builder_name"] = builder_name

        for filetype, flags in self._flags.items():
            flags_dict = {}
            for scope in BuildFlagScope:
                flags_dict[scope.value] = flags[scope]
            data.update({filetype.name: {"flags": flags_dict}})

        return data



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Base source file parser"

import abc
import logging
import os.path as p
from typing import Any, Dict, Optional, Set

from .elements.dependency_spec import DependencySpec  # pylint: disable=unused-import
from .elements.design_unit import tAnyDesignUnit  # pylint: disable=unused-import

from hdl_checker.path import Path
from hdl_checker.types import FileType
from hdl_checker.utils import HashableByKey, removeDuplicates

_logger = logging.getLogger(__name__)


class BaseSourceFile(HashableByKey):  # pylint:disable=too-many-instance-attributes
    """
    Parses and stores information about a source file such as design
    units it depends on and design units it provides
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def _comment(self):
        """
        Should return a regex object that matches a comment (or comments)
        used by the language
        """

    def __init__(self, filename):
        # type: (Path, ) -> None
        assert isinstance(filename, Path), "Invalid type: {}".format(filename)
        self.filename = filename
        self._cache = {}  # type: Dict[str, Any]
        self._content = None
        self._mtime = 0  # type: Optional[float]
        self.filetype = FileType.fromPath(self.filename)
        self._dependencies = None  # type: Optional[Set[DependencySpec]]
        self._design_units = None  # type: Optional[Set[tAnyDesignUnit]]
        self._libraries = None

    def __jsonEncode__(self):
        """
        Gets a dict that describes the current state of this object
        """
        state = self.__dict__.copy()
        del state["_content"]
        del state["_design_units"]
        del state["_dependencies"]
        return state

    @classmethod
    def __jsonDecode__(cls, state):
        """
        Returns an object of cls based on a given state
        """
        obj = super(BaseSourceFile, cls).__new__(cls)
        obj.filename = state["filename"]
        obj.filetype = state["filetype"]
        obj._cache = state["_cache"]  # pylint: disable=protected-access
        obj._content = None  # pylint: disable=protected-access
        obj._mtime = state["_mtime"]  # pylint: disable=protected-access
        obj._dependencies = None  # pylint: disable=protected-access
        obj._design_units = None  # pylint: disable=protected-access
        obj._libraries = state["_libraries"]  # pylint: disable=protected-access

        return obj

    def __repr__(self):
        return "{}(filename={}, design_units={}, dependencies={})".format(
            self.__class__.__name__,
            self.filename,
            self._design_units,
            self._dependencies,
        )

    @property
    def __hash_key__(self):
        return (self.filename, self._content)

    def _changed(self):
        # type: (...) -> Any
        """
        Checks if the file changed based on the modification time
        provided by p.getmtime
        """
        if not p.exists(str(self.filename)):
            return False
        return bool(self.getmtime() > self._mtime)  # type: ignore

    def _clearCachesIfChanged(self):
        # type: () -> None
        """
        Clears all the caches if the file has changed to force updating
        every parsed info
        """
        if self._changed():
            self._content = None
            self._dependencies = None
            self._design_units = None
            self._libraries = None
            self._cache = {}

    def getmtime(self):
        # type: () -> Optional[float]
        """
        Gets file modification time as defined in p.getmtime
        """
        if not p.exists(self.filename.name):
            return None
        return self.filename.mtime

    def getSourceContent(self):
        # type: (...) -> Any
        """
        Cached version of the _getSourceContent method
        """
        self._clearCachesIfChanged()

        if self._content is None:
            self._content = self._getSourceContent()
            self._mtime = self.getmtime()

        return self._content

    def getDesignUnits(self):  # type: () -> Set[tAnyDesignUnit]
        """
        Cached version of the _getDesignUnits method
        """
        if not p.exists(self.filename.name):
            return set()
        self._clearCachesIfChanged()
        if self._design_units is None:
            self._design_units = set(self._getDesignUnits())

        return self._design_units

    def getDependencies(self):
        # type: () -> Set[DependencySpec]
        """
        Cached version of the _getDependencies method
        """
        if not p.exists(self.filename.name):
            return set()

        self._clearCachesIfChanged()
        if self._dependencies is None:
            try:
                self._dependencies = set(self._getDependencies())
            except:  # pragma: no cover
                print("Failed to parse %s" % self.filename)
                _logger.exception("Failed to parse %s", self.filename)
                raise

        return self._dependencies

    def getLibraries(self):
        # type: (...) -> Any
        """
        Cached version of the _getLibraries method
        """
        if not p.exists(self.filename.name):
            return []

        self._clearCachesIfChanged()
        if self._libraries is None:
            self._libraries = removeDuplicates(self._getLibraries())

        return self._libraries

    @abc.abstractmethod
    def _getSourceContent(self):
        """
        Method that should implement pre parsing of the source file.
        This includes removing comments and unnecessary or unimportant
        chunks of text to make the life of the real parsing easier.
        Should return a string and NOT a list of lines
        """

    @abc.abstractmethod
    def _getDesignUnits(self):
        """
        Method that should implement the real parsing of the source file
        to find design units defined. Use the output of the getSourceContent
        method to avoid unnecessary I/O
        """

    @abc.abstractmethod
    def _getLibraries(self):
        """
        Parses the source file to find libraries required by the file
        """

    @abc.abstractmethod
    def _getDependencies(self):
        """
        Parses the source and returns a list of dictionaries that
        describe its dependencies
        """



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined, missing-docstring, protected-access

import logging
import re

import six

from nose2.tools import such  # type: ignore
from nose2.tools.params import params  # type: ignore

from hdl_checker.tests import assertCountEqual

import hdl_checker.static_check as static_check
from hdl_checker.diagnostics import DiagType, LibraryShouldBeOmited, StaticCheckerDiag

_logger = logging.getLogger(__name__)

with such.A("hdl_checker project") as it:
    # Workaround for Python 2.x and 3.x differences
    if six.PY2:
        # Can't use assertCountEqual for lists of unhashable types.
        # Workaround for https://bugs.python.org/issue10242
        it.assertCountEqual = assertCountEqual(it)

    @it.has_setup
    def setup():
        pass

    @it.has_teardown
    def teardown():
        pass

    @it.should("not repeat an object in the results")
    def test():
        text = ["", "library ieee;", "", "library ieee;", ""]

        it.assertDictEqual(
            {"ieee": {"end": 12, "lnum": 1, "start": 8, "type": "library"}},
            static_check._getObjectsFromText(text),
        )

    @it.should("not scan after specific end of scan delimiters")  # type: ignore
    @params(
        [" u0 : some_unit", "   generic map ("],
        [" u0 : some_unit", "   port map ("],
        [" u0 : entity work.some_unit"],
        [" p0 : process"],
        [" process(clk)"],
    )
    def test(case, parm):
        _logger.info("Running test case '%s'", case)
        text = ["library foo;"] + parm + ["library bar;"]

        it.assertDictEqual(
            {"foo": {"end": 11, "lnum": 0, "start": 8, "type": "library"}},
            static_check._getObjectsFromText(text),
        )

    @it.should("extract comment tags")  # type: ignore
    @params(
        " -- XXX: some warning",
        " -- TODO: something to do",
        " -- FIXME: something to fix",
    )
    def test(case, parm):
        _logger.info("Running test case '%s'", case)
        expected = re.sub(r"\s*--\s*", "", parm)

        text = [
            "library ieee;",
            "    use ieee.std_logic_1164.all;",
            "    use ieee.numeric_std.all;",
            "library basic_library;",
            "entity foo is",
            "",
            parm,
            "",
            "    generic (",
            "        DIVIDER_A : integer := 10;",
            "        DIVIDER_B : integer := 20",
            "    );",
            "    port (",
            "        clk_in_a : in std_logic;",
            "        clk_out_a : out std_logic;",
            "",
            "        clk_in_b : in std_logic;",
            "        clk_out_b : out std_logic",
            "",
            "    );",
            "end foo;",
            "",
            "architecture foo of foo is",
            "begin",
            "clk_out_a <= not clk_in_a;",
            "-- clk_out_b <= not clk_in_b;",
            "end architecture foo;",
        ]

        it.assertCountEqual(
            [
                StaticCheckerDiag(
                    line_number=7,
                    column_number=5,
                    severity=DiagType.STYLE_INFO,
                    text=expected,
                )
            ],
            static_check._getCommentTags(text),
        )

    @it.should("get misc checks")  # type: ignore
    def test():
        text = [
            "entity foo is",
            "    port (",
            "        clk_in_a : in std_logic;",
            "        clk_out_a : out std_logic;",
            "",
            "        clk_in_b : in std_logic;",
            "        clk_out_b : out std_logic",
            "",
            "    );",
            "end foo;",
        ]

        objects = static_check._getObjectsFromText(text)

        it.assertCountEqual([], static_check._getMiscChecks(objects))

    with it.having("an entity-architecture pair"):

        @it.has_setup
        def setup():
            it.text = [
                "library ieee;",
                "    use ieee.std_logic_1164.all;",
                "    use ieee.numeric_std.all;",
                "library work;",
                "    use work.some_package.all;",
                "library basic_library;",
                "entity foo is",
                "    generic (",
                "        DIVIDER_A : integer := 10;",
                "        DIVIDER_B : integer := 20",
                "    );",
                "    port (",
                "        clk_in_a : in std_logic;",
                "        clk_out_a : out std_logic;",
                "",
                "        clk_in_b : in std_logic;",
                "        clk_out_b : out std_logic",
                "",
                "    );",
                "end foo;",
                "",
                "architecture foo of foo is",
                "begin",
                "clk_out_a <= not clk_in_a;",
                "-- clk_out_b <= not clk_in_b;",
                "end architecture foo;",
            ]

        @it.should("get VHDL objects from an entity-architecture pair")  # type: ignore
        def test():
            it.assertDictEqual(
                {
                    "DIVIDER_A": {"end": 18, "lnum": 8, "start": 8, "type": "generic"},
                    "DIVIDER_B": {"end": 18, "lnum": 9, "start": 8, "type": "generic"},
                    "basic_library": {
                        "end": 21,
                        "lnum": 5,
                        "start": 8,
                        "type": "library",
                    },
                    "clk_in_a": {"end": 17, "lnum": 12, "start": 8, "type": "port"},
                    "clk_in_b": {"end": 17, "lnum": 15, "start": 8, "type": "port"},
                    "clk_out_a": {"end": 18, "lnum": 13, "start": 8, "type": "port"},
                    "clk_out_b": {"end": 18, "lnum": 16, "start": 8, "type": "port"},
                    "ieee": {"end": 12, "lnum": 0, "start": 8, "type": "library"},
                    "work": {"end": 12, "lnum": 3, "start": 8, "type": "library"},
                },
                static_check._getObjectsFromText(it.text),
            )

        @it.should("get misc checks")  # type: ignore
        def test():
            objects = static_check._getObjectsFromText(it.text)

            it.assertCountEqual(
                [LibraryShouldBeOmited(line_number=4, column_number=9, library="work")],
                static_check._getMiscChecks(objects),
            )

        @it.should("get unused VHDL objects")  # type: ignore
        def test():
            objects = static_check._getObjectsFromText(it.text)
            it.assertCountEqual(
                ["basic_library", "DIVIDER_A", "DIVIDER_B", "clk_in_b", "clk_out_b"],
                static_check._getUnusedObjects(it.text, objects),
            )

    with it.having("a package-package body pair"):

        @it.has_setup
        def setup():
            it.text = [
                "library ieee;",
                "package very_common_pkg is",
                '    constant VIM_HDL_VERSION : string := "0.1";',
                "end package;",
                "package body very_common_pkg is",
                "end package body;",
            ]

        @it.should("get VHDL objects from a package-package body pair")  # type: ignore
        def test():

            it.assertDictEqual(
                {"ieee": {"end": 12, "lnum": 0, "start": 8, "type": "library"}},
                static_check._getObjectsFromText(it.text),
            )

        @it.should("get unused VHDL objects")  # type: ignore
        def test():
            objects = static_check._getObjectsFromText(it.text)
            it.assertCountEqual(
                ["ieee"], static_check._getUnusedObjects(it.text, objects)
            )


it.createTests(globals())



#!/usr/bin/env python
# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"HDL Checker server launcher"

# PYTHON_ARGCOMPLETE_OK

import argparse
import logging
import os
import sys
from threading import Timer

from pyls.python_ls import start_io_lang_server  # type: ignore

from hdl_checker import __version__ as version
from hdl_checker import handlers, lsp
from hdl_checker.utils import (
    getTemporaryFilename,
    isProcessRunning,
    setupLogging,
    terminateProcess,
)

_logger = logging.getLogger(__name__)
PY2 = sys.version_info[0] == 2

_LSP_ERROR_MSG_TEMPLATE = {"method": "window/showMessage", "jsonrpc": "2.0"}


def parseArguments():
    "Argument parser for standalone hdl_checker"

    parser = argparse.ArgumentParser()

    # Options
    parser.add_argument("--host", action="store", help="[HTTP] Host to serve")
    parser.add_argument("--port", action="store", type=int, help="[HTTP] Port to serve")
    parser.add_argument(
        "--lsp",
        action="store_true",
        default=False,
        help="Starts the server in LSP mode. Defaults to false",
    )

    parser.add_argument(
        "--attach-to-pid",
        action="store",
        type=int,
        help="[HTTP, LSP] Stops the server if given PID is not active",
    )
    parser.add_argument("--log-level", action="store", help="[HTTP, LSP] Logging level")
    parser.add_argument(
        "--log-stream",
        action="store",
        help="[HTTP, LSP] Log file, defaults to stdout when in HTTP or a "
        "temporary file named hdl_checker_log_pid<PID>.log when in LSP mode",
    )

    parser.add_argument(
        "--stdout",
        action="store",
        help="[HTTP] File to redirect stdout to. Defaults to a temporary file "
        "named hdl_checker_stdout_pid<PID>.log",
    )
    parser.add_argument(
        "--stderr",
        action="store",
        help="[HTTP] File to redirect stdout to. Defaults to a temporary file "
        "named hdl_checker_stderr_pid<PID>.log",
    )

    parser.add_argument(
        "--version",
        "-V",
        action="store_true",
        help="Prints hdl_checker version and exit",
    )

    try:
        import argcomplete  # type: ignore

        argcomplete.autocomplete(parser)
    except ImportError:  # pragma: no cover
        pass

    args = parser.parse_args()

    if args.version:
        sys.stdout.write("%s\n" % version)
        sys.exit(0)

    if args.lsp:
        args.host = None
        args.port = None
    else:
        args.host = args.host or "localhost"
        args.port = args.port or 50000
        args.log_stream = args.log_stream or sys.stdout

    # If not set, create a temporary file safely so there's no clashes
    args.log_stream = args.log_stream or getTemporaryFilename("log")
    args.stderr = args.stderr or getTemporaryFilename("stderr")

    args.log_level = args.log_level or logging.INFO

    return args


# Copied from ycmd!
def openForStdHandle(filepath):
    """
    Returns a file object that can be used to replace sys.stdout or
    sys.stderr
    """
    # Need to open the file in binary mode on py2 because of bytes vs unicode.
    # If we open in text mode (default), then third-party code that uses `print`
    # (we're replacing sys.stdout!) with an `str` object on py2 will cause
    # tracebacks because text mode insists on unicode objects. (Don't forget,
    # `open` is actually `io.open` because of future builtins.)
    # Since this function is used for logging purposes, we don't want the output
    # to be delayed. This means no buffering for binary mode and line buffering
    # for text mode. See https://docs.python.org/2/library/io.html#io.open
    if PY2:
        return open(filepath, mode="wb", buffering=0)
    return open(filepath, mode="w", buffering=1)


def _setupPipeRedirection(stdout, stderr):  # pragma: no cover
    "Redirect stdout and stderr to files"
    if stdout is not None:
        sys.stdout = openForStdHandle(stdout)
    if stderr is not None:
        sys.stderr = openForStdHandle(stderr)


def _binaryStdio():
    """
    (from https://github.com/palantir/python-language-server)

    This seems to be different for Window/Unix Python2/3, so going by:
        https://stackoverflow.com/questions/2850893/reading-binary-data-from-stdin
    """

    if not PY2:
        # pylint: disable=no-member
        stdin, stdout = sys.stdin.buffer, sys.stdout.buffer
    else:
        # Python 2 on Windows opens sys.stdin in text mode, and
        # binary data that read from it becomes corrupted on \r\n
        if sys.platform == "win32":
            # set sys.stdin to binary mode
            # pylint: disable=no-member,import-error
            import msvcrt

            msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
            msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)
        stdin, stdout = sys.stdin, sys.stdout

    return stdin, stdout


def run(args):  # pylint: disable=missing-docstring
    try:
        # LSP will use stdio to communicate
        _setupPipeRedirection(None if args.lsp else args.stdout, args.stderr)

        startServer(args)
    except Exception as exc:
        if args.lsp:  # pragma: no cover
            msg = ["Unable to start HDL Checker LSP server: %s!" % repr(exc)]
            if args.stderr:
                msg += ["Check %s for more info" % args.stderr]
            else:
                msg += ["Use --stderr to redirect the output to a file for more info"]
            _reportException(" ".join(msg))
        _logger.exception("Failed to start server")
        raise


def startServer(args):
    """
    Import modules and tries to start a hdl_checker server
    """
    if args.log_stream:
        setupLogging(args.log_stream, args.log_level)

    globals()["_logger"] = logging.getLogger(__name__)

    def _attachPids(source_pid, target_pid):
        "Monitors if source_pid is alive. If not, terminate target_pid"

        def _watchPidWrapper():
            "PID attachment monitor"
            try:
                if isProcessRunning(source_pid):
                    Timer(1, _watchPidWrapper).start()
                else:
                    _logger.warning("Process %d is not running anymore", source_pid)
                    terminateProcess(target_pid)
            except (TypeError, AttributeError):  # pragma: no cover
                return

        _logger.debug("Setting up PID attachment from %s to %s", source_pid, target_pid)

        Timer(2, _watchPidWrapper).start()

    _logger.info(
        "Starting server. Our PID is %s, %s. Version string for hdl_checker is '%s'",
        os.getpid(),
        "no parent PID to attach to"
        if args.attach_to_pid is None
        else "our parent is %s" % args.attach_to_pid,
        version,
    )

    if args.lsp:
        stdin, stdout = _binaryStdio()
        start_io_lang_server(stdin, stdout, True, lsp.HdlCheckerLanguageServer)
    else:
        if args.attach_to_pid is not None:
            _attachPids(args.attach_to_pid, os.getpid())

        handlers.app.run(host=args.host, port=args.port, threads=10, server="waitress")


# This is a redefinition to be used as last resort if failed to setup
# the server
class MessageType:  # pylint: disable=too-few-public-methods
    "LSP message types"
    Error = 1
    Warning = 2
    Info = 3
    Log = 4


def _reportException(text):
    "Hand crafted message to report issues when starting up the HDL Checker server"
    _logger.warning("Reporting exception with text %s", text)

    import json

    message = _LSP_ERROR_MSG_TEMPLATE.copy()
    message["params"] = {"message": text, "type": MessageType.Error}

    body = json.dumps(message)

    # Ensure we get the byte length, not the character length
    content_length = len(body) if isinstance(body, bytes) else len(body.encode("utf-8"))
    response = (
        "Content-Length: {}\r\n"
        "Content-Type: application/vscode-jsonrpc; charset=utf8\r\n\r\n"
        "{}".format(content_length, body)
    )

    sys.stdout.write(response)


def main():
    return run(parseArguments())


if __name__ == "__main__":
    main()



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Path helper class to speed up comparing different paths"

# pylint: disable=useless-object-inheritance

import logging
from os import path as p
from os import stat
from typing import Union

import six

_logger = logging.getLogger(__name__)

if six.PY2:
    FileNotFoundError = OSError  # pylint: disable=redefined-builtin


class Path(object):
    "Path helper class to speed up comparing different paths"

    def __init__(self, name, base_path=None):
        # type: (Union[Path, str], Union[Path, str, None]) -> None
        assert isinstance(
            name, (Path, six.string_types)
        ), "Invalid type for path: {} ({})".format(name, type(name))

        if p.isabs(str(name)) or base_path is None:
            _name = name
        else:
            _name = p.join(str(base_path), str(name))
        self._name = p.normpath(str(_name))

    @property
    def mtime(self):
        # type: () -> float
        """
        Equivalent to os.path.getmtime(self.name)
        """
        return p.getmtime(self.name)

    @property
    def abspath(self):
        # type: () -> str
        """
        Equivalent to os.path.abspath(self.name)
        """
        return p.abspath(self.name)

    @property
    def basename(self):
        # type: () -> str
        """
        Equivalent to os.path.basename(self.name)
        """
        return p.basename(self.name)

    @property
    def dirname(self):
        # type: () -> str
        """
        Equivalent to os.path.dirname(self.name)
        """
        return p.dirname(self.name)

    @property
    def name(self):
        """
        Absolute path, either the path passed to the constructor or the path
        prepended with base_path. In the second case, it's up to the caller to
        ensure an absolute path can be constructed; no exception or warning is
        thrown.
        """
        return self._name

    def __str__(self):
        return self.name

    def __repr__(self):
        return "Path({})".format(repr(self.name))

    @property
    def stat(self):
        """
        Equivalent to os.path.stat(self.name)
        """
        return stat(self.name)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        """Overrides the default implementation"""
        try:
            # Same absolute paths mean the point to the same file. Prefer this
            # to avoid calling os.stat all the time
            if self.abspath == other.abspath:
                return True
            return p.samestat(self.stat, other.stat)
        except (AttributeError, FileNotFoundError):
            return False

        return NotImplemented  # pragma: no cover

    def __ne__(self, other):  # pragma: no cover
        """Overrides the default implementation (unnecessary in Python 3)"""
        result = self.__eq__(other)

        if result is NotImplemented:
            return NotImplemented

        return not result

    def __jsonEncode__(self):
        """
        Gets a dict that describes the current state of this object
        """
        return {"name": self.name}

    @classmethod
    def __jsonDecode__(cls, state):
        """Returns an object of cls based on a given state"""

        obj = super(Path, cls).__new__(cls)
        obj._name = state["name"]  # pylint: disable=protected-access

        return obj


class TemporaryPath(Path):
    """
    Class made just to differentiate a path from a temporary path created to
    dump a source's content
    """



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"hdl_checker installation script"

import setuptools
import versioneer

LONG_DESCRIPTION = open("README.md", "rb").read().decode(encoding='utf8', errors='replace')

CLASSIFIERS = """\
Development Status :: 5 - Production/Stable
Environment :: Console
Intended Audience :: Developers
License :: OSI Approved :: GNU General Public License v3 (GPLv3)
Operating System :: Microsoft :: Windows
Operating System :: POSIX :: Linux
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 3
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Topic :: Software Development
Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)
Topic :: Text Editors :: Integrated Development Environments (IDE)
"""

# pylint: disable=bad-whitespace
setuptools.setup(
    name                          = 'hdl_checker',
    version                       = versioneer.get_version(),
    description                   = 'HDL code checker',
    long_description              = LONG_DESCRIPTION,
    long_description_content_type = "text/markdown",
    author                        = 'Andre Souto',
    author_email                  = 'andre820@gmail.com',
    url                           = 'https://github.com/suoto/hdl_checker',
    license                       = 'GPLv3',
    keywords                      = 'VHDL Verilog SystemVerilog linter LSP language server protocol vimhdl vim-hdl',
    platforms                     = 'any',
    packages                      = setuptools.find_packages(),
    install_requires              = ['argcomplete',
                                     'argparse',
                                     'backports.functools_lru_cache; python_version<"3.2"',
                                     'bottle>=0.12.9',
                                     'enum34>=1.1.6; python_version<"3.3"',
                                     'future>=0.14.0',
                                     'futures; python_version<"3.2"',
                                     'prettytable>=0.7.2',
                                     'python-language-server>=0.26.1,<=0.27.0',
                                     'requests==2.20.0',
                                     'tabulate>=0.8.5',
                                     'typing>=3.7.4',
                                     'waitress>=0.9.0', ],
    cmdclass                      = versioneer.get_cmdclass(),
    entry_points                  = {
        'console_scripts' : ['hdl_checker=hdl_checker.server:main', ]
    },
    classifiers=CLASSIFIERS.splitlines(),
)
# pylint: enable=bad-whitespace



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"VHDL source file parser"

import logging
import re
from typing import Any, Generator

from .elements.design_unit import VerilogDesignUnit
from .elements.parsed_element import Location

from hdl_checker.parsers.base_parser import BaseSourceFile
from hdl_checker.types import DesignUnitType
from hdl_checker.utils import readFile

_logger = logging.getLogger(__name__)

_VERILOG_IDENTIFIER = r"[a-zA-Z_][a-zA-Z0-9_$]+"
# Design unit scanner
_DESIGN_UNIT_SCANNER = re.compile(
    "|".join(
        [
            r"\bmodule\s+(?P<module_name>%s)" % _VERILOG_IDENTIFIER,
            r"\bpackage\s+(?P<package_name>%s)" % _VERILOG_IDENTIFIER,
        ]
    ),
    flags=re.S,
)


class VerilogParser(BaseSourceFile):
    """
    Parses and stores information about a Verilog or SystemVerilog
    source file
    """

    _comment = re.compile(r"\/\*.*?\*\/|//[^(\r\n?|\n)]*", flags=re.DOTALL)

    def _getSourceContent(self):
        # type: (...) -> Any
        # Remove multiline comments
        content = readFile(str(self.filename))
        lines = self._comment.sub("", content)
        return re.sub(r"\r\n?|\n", " ", lines, flags=re.S)

    def _iterDesignUnitMatches(self):
        # type: (...) -> Any
        """
        Iterates over the matches of _DESIGN_UNIT_SCANNER against
        source's lines
        """
        content = self.getSourceContent()
        for match in _DESIGN_UNIT_SCANNER.finditer(self.getSourceContent()):
            start = match.start()
            yield match.groupdict(), content[:start].count("\n")

    def _getDependencies(self):
        return []

    def _getDesignUnits(self):  # type: () -> Generator[VerilogDesignUnit, None, None]
        for match, line_number in self._iterDesignUnitMatches():
            if match["module_name"] is not None:
                yield VerilogDesignUnit(
                    owner=self.filename,
                    name=match["module_name"],
                    type_=DesignUnitType.entity,
                    locations={(line_number, None)},
                )
            if match["package_name"] is not None:
                yield VerilogDesignUnit(
                    owner=self.filename,
                    name=match["package_name"],
                    type_=DesignUnitType.package,
                    locations={(line_number, None)},
                )

    def _getLibraries(self):
        return set()
